package au.gov.health.covidsafe.extensions

import android.Manifest.permission.ACCESS_COARSE_LOCATION
import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothManager
import android.content.Context
import android.content.Intent
import android.location.LocationManager
import android.os.Build
import android.os.PowerManager
import android.provider.Settings
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.core.app.NotificationManagerCompat
import androidx.fragment.app.Fragment
import au.gov.health.covidsafe.R
import au.gov.health.covidsafe.Utils
import pub.devrel.easypermissions.EasyPermissions
import pub.devrel.easypermissions.PermissionRequest


const val REQUEST_ENABLE_BT = 123
const val LOCATION = 345
const val BATTERY_OPTIMISER = 789

fun Fragment.requestAllPermissions(onEndCallback: () -> Unit) {
    if (isBlueToothEnabled() ?: true) {
        requestFineLocationAndCheckBleSupportThenNextPermission(onEndCallback)
    } else {
        requestBlueToothPermissionThenNextPermission()
    }
}

fun Fragment.requestBlueToothPermissionThenNextPermission() {
    val enableBtIntent = Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE)
    startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT)
}

fun Fragment.checkBLESupport() {
    if (BluetoothAdapter.getDefaultAdapter()?.isMultipleAdvertisementSupported?.not() ?: false) {
        activity?.let {
            Utils.stopBluetoothMonitoringService(it)
        }
    }
}

private fun Fragment.requestFineLocationAndCheckBleSupportThenNextPermission(onEndCallback: () -> Unit) {
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
        activity?.let {
            when {
                EasyPermissions.hasPermissions(it, ACCESS_COARSE_LOCATION) -> {
                    checkBLESupport()
                    excludeFromBatteryOptimization(onEndCallback)
                }
                else -> {
                    EasyPermissions.requestPermissions(
                            PermissionRequest.Builder(this, LOCATION, ACCESS_COARSE_LOCATION)
                                    .setRationale(R.string.permission_location_rationale)
                                    .build())
                }
            }
        }
    } else {
        checkBLESupport()
    }
}

fun Fragment.excludeFromBatteryOptimization(onEndCallback: (() -> Unit)? = null) {
    activity?.let {
        val powerManager =
                it.getSystemService(AppCompatActivity.POWER_SERVICE) as PowerManager
        val packageName = it.packageName
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            val intent = Utils.getBatteryOptimizerExemptionIntent(packageName)
            if (!powerManager.isIgnoringBatteryOptimizations(packageName)) {
                //check if there's any activity that can handle this
                if (Utils.canHandleIntent(intent, it.packageManager)) {
                    this.startActivityForResult(intent, BATTERY_OPTIMISER)
                } else {
                    //no way of handling battery optimizer
                    onEndCallback?.invoke()
                }
            } else {
                onEndCallback?.invoke()
            }
        }
    }

}

fun Fragment.gotoPushNotificationSettings() {
    val context = requireContext()
    val intent = Intent()

    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
        intent.action = Settings.ACTION_APP_NOTIFICATION_SETTINGS
        intent.putExtra(Settings.EXTRA_APP_PACKAGE, context.packageName)
    } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
        intent.action = "android.settings.APP_NOTIFICATION_SETTINGS"
        intent.putExtra("app_package", context.packageName)
        intent.putExtra("app_uid", context.applicationInfo.uid)
    }

    context.startActivity(intent)
}

fun Fragment.isBlueToothEnabled(): Boolean? {
    val bluetoothManager = activity?.getSystemService(Context.BLUETOOTH_SERVICE) as BluetoothManager?
    return bluetoothManager?.adapter?.isEnabled
}

fun Fragment.isPushNotificationEnabled(): Boolean? {
    return activity?.let { activity ->
        NotificationManagerCompat.from(activity).areNotificationsEnabled()
    }
}

fun Fragment.isLocationPermissionAllowed(): Boolean? {
    return activity?.let { activity ->
        EasyPermissions.hasPermissions(activity, ACCESS_COARSE_LOCATION)
    }
}

fun Fragment.isLocationEnabledOnDevice(): Boolean {
    val locationManager = context?.getSystemService(Context.LOCATION_SERVICE) as LocationManager?
    return locationManager?.let {
        it.isProviderEnabled(LocationManager.GPS_PROVIDER) ||
                it.isProviderEnabled(LocationManager.NETWORK_PROVIDER)
    } ?: false
}

fun Fragment.isBatteryOptimizationDisabled(): Boolean? {
    return activity?.let { activity ->
        val powerManager = activity.getSystemService(AppCompatActivity.POWER_SERVICE) as PowerManager?
        val packageName = activity.packageName

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            powerManager?.isIgnoringBatteryOptimizations(packageName) ?: true
        } else {
            null
        }
    } ?: run {
        null
    }
}

fun Fragment.askForLocationPermission() {
    activity?.let {
        when {
            !EasyPermissions.hasPermissions(it, ACCESS_COARSE_LOCATION) -> {
                EasyPermissions.requestPermissions(
                        PermissionRequest.Builder(this, LOCATION, ACCESS_COARSE_LOCATION)
                                .setRationale(R.string.permission_location_rationale)
                                .build())
            }

            !isLocationEnabledOnDevice() -> {
                AlertDialog.Builder(it).apply {
                    setMessage(R.string.need_location_service)
                    setPositiveButton(android.R.string.ok) { _, _ ->
                        it.startActivity(Intent(Settings.ACTION_LOCATION_SOURCE_SETTINGS))
                    }
                }.create().show()
            }

            else -> {
            }// do nothing
        }
    }
}
